import dataclasses
import re
from dataclasses import dataclass, field
from typing import Optional, Dict

from openpyxl.cell import Cell
from openpyxl.styles import Border, Side, Alignment, Font
from openpyxl.styles.alignment import vertical_aligments
from requests_html import Element

from jinja2xlsx.utils import union_dicts, CellRange

REMOVE_SIDE = Side()


@dataclass()
class Style:
    border: Border = field(default_factory=Border)
    alignment: Alignment = field(default_factory=Alignment)
    font: Font = field(default_factory=Font)

    def union(self, style: 'Style') -> 'Style':
        """
        >>> from openpyxl.styles import Side
        >>> default_style = Style(alignment=Alignment(wrap_text=True), font=Font("Times New Roman", 10))
        >>> style = Style(border=Border(left=Side("medium")), font=Font(sz=11, bold=True))
        >>> new_style = default_style.union(style)
        >>> new_style.border.left.style
        'medium'
        >>> new_style.alignment.wrap_text
        True
        >>> new_style.font == Font("Times New Roman", 11, bold=True)
        True
        """
        border_data = union_dicts(vars(self.border), vars(style.border))
        alignment_data = union_dicts(vars(self.alignment), vars(style.alignment))
        font_data = union_dicts(vars(self.font), vars(style.font))

        return Style(Border(**border_data), Alignment(**alignment_data), Font(**font_data))


def extract_style(style_attr: str) -> Style:
    """
    >>> style = extract_style("border: 1px solid black; text-align: center; font-weight: bold")
    >>> style.alignment.horizontal
    'center'
    >>> style.border.left.style
    'thin'
    >>> style.border.left.style == style.border.right.style == style.border.top.style == style.border.bottom.style
    True
    >>> style.font.bold
    True
    """
    if not style_attr:
        return Style()

    style_dict = parse_style_attr(style_attr)

    border = _build_border(style_dict)
    alignment = _build_alignment(style_dict)
    font = _build_font(style_dict)

    return Style(border, alignment, font)


@dataclass()
class Stylist:
    default_style: 'Style' = field(default_factory=Style)

    def build_style_from_html(self, html_element: Element) -> Style:
        style_attr = html_element.attrs.get("style")
        style = extract_style(style_attr)
        style.font.bold = style.font.bold or html_element.tag == "th"
        style = self.default_style.union(style)
        return style

    def style_single_cell(self, cell: Cell, style: Style) -> None:
        cell.border = style.border
        cell.alignment = style.alignment
        cell.font = style.font

    def style_merged_cells(self, cell_range: CellRange, style: Style) -> None:
        """
        Source:
        https://openpyxl.readthedocs.io/en/2.5/styles.html#styling-merged-cells
        """
        first_cell = cell_range[0][0]
        first_cell.alignment = style.alignment
        first_cell.font = style.font

        top = Border(top=style.border.top)
        left = Border(left=style.border.left)
        right = Border(right=style.border.right)
        bottom = Border(bottom=style.border.bottom)

        for cell in cell_range[0]:
            cell.border = cell.border + top
        for cell in cell_range[-1]:
            cell.border = cell.border + bottom

        for row in cell_range:
            left_cell = row[0]
            right_cell = row[-1]
            left_cell.border = left_cell.border + left
            right_cell.border = right_cell.border + right


def parse_style_attr(style_str: Optional[str]) -> Dict:
    """
    >>> parse_style_attr("border: 1px solid black; text-align: center; font-weight: bold")
    {'border': '1px solid black', 'text-align': 'center', 'font-weight': 'bold'}
    >>> parse_style_attr("")
    {}
    >>> parse_style_attr(None)
    {}
    >>> parse_style_attr('border: 1px solid black; ')
    {'border': '1px solid black'}
    """
    if not style_str:
        return {}

    try:
        return {
            style.strip(): value.strip()
            for style, value in (style.split(":") for style in filter(None, map(str.strip, style_str.split(";"))))
        }
    except ValueError:
        raise ValueError(f'Failed to parse style: {style_str}')


@dataclasses.dataclass
class ParseBorder:
    style_dict: Dict[str, str]

    def __call__(self) -> Border:
        final_border = Border()

        border_style_attrs = [
            'border',
            'border-left',
            'border-right',
            'border-top',
            'border-bottom',
        ]
        border_style_attrs = [attr for attr in border_style_attrs if attr in self.style_dict]

        for attr in border_style_attrs:
            if attr == 'border':
                final_border.left = self._parse_b_value(self.style_dict[attr])
                final_border.right = self._parse_b_value(self.style_dict[attr])
                final_border.top = self._parse_b_value(self.style_dict[attr])
                final_border.bottom = self._parse_b_value(self.style_dict[attr])
            else:
                side = attr.split('-')[1]
                setattr(final_border, side, self._parse_b_value(self.style_dict[attr]))

        return final_border

    def _parse_b_value(self, b_value: str) -> Side:
        if b_value == '1px solid black':
            return Side('thin')
        if re.match(r'\d+px solid black', b_value):
            return Side('medium')
        if b_value == '0' or b_value == 'none':
            return Side()


def _build_border(style_dict: Dict[str, str]) -> Border:
    """
    >>> border = _build_border({"border": "1px solid black"})
    >>> border.left.style
    'thin'
    >>> border.left.style == border.right.style == border.top.style == border.bottom.style
    True
    >>> border = _build_border({"border-right": "2px solid black"})
    >>> border.right.style
    'medium'
    >>> border = _build_border({"border": "1px solid black", "border-bottom": "0"})
    >>> border.left == border.right == border.top == Side("thin")
    True
    >>> border.bottom == Side()
    True
    >>> border = _build_border({"border": "1px solid black", "border-top": "none"})
    >>> border.left == border.right == border.bottom == Side("thin")
    True
    >>> border.top == Side()
    True
    """
    return ParseBorder(style_dict)()


def _build_alignment(style_dict: Dict) -> Alignment:
    """
    >>> _build_alignment({'vertical-align': 'super'})
    Traceback (most recent call last):
    ...
    AssertionError: vertical-align should be in ('top', 'center', 'bottom', 'justify', 'distributed'), got: super
    """
    h_align = style_dict.get("text-align")
    # html-v-align: baseline | sub | super | text-top | text-bottom | middle | top | bottom | <percentage> | <length>
    # xlsx-v-align: "top", "center", "bottom", "justify", "distributed",
    v_align = style_dict.get('vertical-align')
    if v_align:
        if v_align == 'middle':
            v_align = 'center'
        assert v_align in vertical_aligments, \
            f'vertical-align should be in {vertical_aligments}, got: {v_align}'


    wrap_text: Optional[bool]
    word_wrap = style_dict.get("word-wrap")
    if word_wrap == "break-word":
        wrap_text = True
    elif word_wrap == "normal":
        wrap_text = False
    else:
        wrap_text = None

    alignment = Alignment(horizontal=h_align, vertical=v_align, wrap_text=wrap_text)
    return alignment


def _build_font(style_dict: Dict) -> Font:
    """
    >>> _build_font({'font-size': '8px'}).size == 8
    True
    """
    bold = style_dict.get("font-weight") == "bold"

    size: Optional[int] = None
    size_str = style_dict.get('font-size')
    if size_str:
        size = int(re.findall(r'(\d+)px', size_str)[0])

    return Font(
        bold=bold,
        size=size,
    )



from dataclasses import dataclass
from typing import Sequence, Optional

from cached_property import cached_property
from requests_html import HTML, Element


@dataclass()
class Parser:
    html_str: str

    @cached_property
    def html(self) -> HTML:
        return HTML(html=self.html_str)

    @cached_property
    def table(self) -> Element:
        table = self.html.find("table", first=True)
        assert table
        return table

    @cached_property
    def table_head(self) -> Element:
        thead = self.table.find("thead", first=True)
        return thead

    @cached_property
    def table_body(self) -> Element:
        tbody = self.table.find("tbody", first=True)
        assert tbody
        return tbody

    @cached_property
    def rows(self) -> Sequence[Element]:
        return [
            *(self.table_head.find("tr") if self.table_head else []),
            *(self.table_body.find("tr")),
        ]

    @cached_property
    def colgroup(self) -> Optional[Element]:
        return self.table.find("colgroup", first=True)

    @cached_property
    def columns(self) -> Sequence[Element]:
        return self.colgroup.find("col") if self.colgroup else []



from dataclasses import dataclass
from typing import Iterable

from openpyxl.utils import get_column_letter
from openpyxl.worksheet.dimensions import ColumnDimension
from openpyxl.worksheet.worksheet import Worksheet
from requests_html import Element

from jinja2xlsx.style import parse_style_attr
from jinja2xlsx.utils import (
    try_extract_pixels,
    width_pixels_to_xlsx_units,
    height_pixels_to_xlsx_units,
)


@dataclass
class Adjuster:
    sheet: Worksheet

    def adjust_columns(self, columns: Iterable[Element]) -> None:
        for index, column in enumerate(columns):
            col_width_in_pixels = int(column.attrs.get("width", 0))
            if not col_width_in_pixels:
                continue

            column_dimension: ColumnDimension = self.sheet.column_dimensions[
                get_column_letter(index + 1)
            ]
            column_dimension.width = width_pixels_to_xlsx_units(col_width_in_pixels)

    def adjust_rows(self, rows: Iterable[Element]) -> None:
        for index, row in enumerate(rows):
            # todo there must be a better way
            style_dict = parse_style_attr(row.attrs.get("style"))
            height_str = style_dict.get("line-height") or style_dict.get("height") or ""
            row_height = try_extract_pixels(height_str)
            if not row_height:
                continue

            self.sheet.row_dimensions[index + 1].height = height_pixels_to_xlsx_units(row_height)



import os
from dataclasses import dataclass
from typing import Optional

BASE_DIR = os.path.dirname(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(BASE_DIR, "test_data")


@dataclass
class Config:
    # if td contains img tag => get base64 src from img and insert to xlsx
    parse_img: bool = False
    # get url src from img, download it via requests and insert to xlsx
    parse_img_url: bool = False
    # required for relative image url like /img/cat.jpg
    base_url: Optional[str] = None



from dataclasses import dataclass
from typing import Tuple

from openpyxl import Workbook
from openpyxl.cell import MergedCell, Cell
from openpyxl.worksheet.worksheet import Worksheet

from jinja2xlsx.adjust import Adjuster
from jinja2xlsx.config import Config
from jinja2xlsx.image import ImageParse
from jinja2xlsx.parse import Parser
from jinja2xlsx.style import Stylist
from jinja2xlsx.utils import CellGenerator, create_cell_range_str, parse_cell_value


@dataclass
class Renderer:
    parser: Parser
    stylist: Stylist
    config: Config
    workbook: Workbook = None
    sheet: Worksheet = None

    def __call__(self) -> Workbook:
        self.wb = Workbook()
        self.sheet = self.wb.active

        cells = list(self._generate_cells())
        self._fill_cells(cells)
        self._style_cells(cells)

        adjuster = Adjuster(self.sheet)
        adjuster.adjust_columns(self.parser.columns)
        adjuster.adjust_rows(self.parser.rows)

        return self.wb

    def _generate_cells(self) -> CellGenerator:
        for row_index, row in enumerate(self.parser.rows):
            col_index = 0

            html_cells = [*row.find("th"), *row.find("td")]
            for html_cell in html_cells:
                target_cell, col_index = self._find_free_cell(col_index, row_index)

                colspan = int(html_cell.attrs.get("colspan", 1))
                rowspan = int(html_cell.attrs.get("rowspan", 1))

                if colspan > 1 or rowspan > 1:
                    cell_range_str = create_cell_range_str(col_index, colspan, row_index, rowspan)
                    self.sheet.merge_cells(cell_range_str)
                    yield html_cell, None, self.sheet[cell_range_str]
                else:
                    yield html_cell, target_cell, None

                col_index += colspan

    def _find_free_cell(self, col_index: int, row_index: int) -> Tuple[Cell, int]:
        target_cell = self.sheet.cell(row_index + 1, col_index + 1)

        while True:
            if isinstance(target_cell, MergedCell):
                col_index += 1
                target_cell = self.sheet.cell(row_index + 1, col_index + 1)
            else:
                break

        return target_cell, col_index

    def _fill_cells(self, cells: CellGenerator) -> None:
        for html_cell, cell, cell_range in cells:
            target_cell = None
            if cell:
                target_cell = cell
            if cell_range:
                target_cell = cell_range[0][0]
            assert target_cell

            image_tag = html_cell.find("img", first=True)
            if image_tag and self.config.parse_img:
                image = ImageParse(self.config)(image_tag)
                self.sheet.add_image(image, target_cell.coordinate)
            else:
                target_cell.value = parse_cell_value(html_cell.text)

    def _style_cells(self, cells: CellGenerator) -> None:
        for html_cell, cell, cell_range in cells:
            style = self.stylist.build_style_from_html(html_cell)

            if cell:
                self.stylist.style_single_cell(cell, style)
            elif cell_range:
                self.stylist.style_merged_cells(cell_range, style)



import base64
import io
import re
from dataclasses import dataclass
from typing import Optional
from urllib.parse import urljoin

import requests
from openpyxl.drawing.image import Image
from requests_html import Element

from jinja2xlsx.config import Config


@dataclass
class ImageParse:
    config: Config

    def __call__(self, image_tag: Element) -> Image:
        src = image_tag.attrs["src"]

        base64 = try_base64(src)
        if base64:
            image_stream: io.BytesIO = base64_to_stream(base64)
        elif is_url(src):
            image_stream = io.BytesIO(requests.get(src).content)
        elif self.config.base_url:
            src = urljoin(self.config.base_url, src)
            image_stream = io.BytesIO(requests.get(src).content)
        else:
            raise ValueError(f"No [Config.base_url] set, so cannot resolve image src: {src}")

        return Image(image_stream)


def is_url(src: str) -> bool:
    """
    >>> is_url("http://www.example.com/image.gif")
    True
    >>> is_url("data:image/png;base64, iVBORw0KGgoAAAANSUhEUgAAAAUA\\nAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO\\n9TXL0Y4OHwAAAABJRU5ErkJggg==")
    False
    >>> is_url("smiley.gif")
    False

    https://stackoverflow.com/a/7160778/5500609
    """

    regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$',
        re.IGNORECASE,
    )
    return re.match(regex, src) is not None


def try_base64(src: str) -> Optional[str]:
    """
    >>> try_base64("data:image/png;base64, iVBORw0KGgoAAAANSUhEUgAAAAUA\\nAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO\\n9TXL0Y4OHwAAAABJRU5ErkJggg==")
    'iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg=='
    >>> try_base64("smiley.gif") is None
    True
    >>> try_base64("http://www.example.com/image.gif") is None
    True
    """
    try:
        base64_str = re.findall(r"data:.*?;base64,([\s\S]*)", src)[0]
        base64_str = base64_str.strip().replace("\n", "")
        return base64_str
    except IndexError:
        return None


def base64_to_stream(base64_str: str) -> io.BytesIO:
    return io.BytesIO(base64.b64decode(base64_str))



import os
from contextlib import contextmanager
from typing import List, Tuple, Iterator, TextIO

from openpyxl import Workbook

from jinja2xlsx.config import TEST_DATA_DIR


def get_test_file_path(file_: str) -> str:
    return os.path.join(TEST_DATA_DIR, file_)


@contextmanager
def read_from_test_dir(file_: str) -> Iterator[TextIO]:
    with open(get_test_file_path(file_), encoding="utf-8") as f:
        yield f


def get_wb_values(wb: Workbook) -> List[Tuple]:
    return list(wb.active.values)



import re
from typing import Dict, Optional, Any, Tuple, Iterable

from openpyxl.cell import Cell
from openpyxl.utils import get_column_letter
from requests_html import Element

CellRange = Tuple[Tuple[Cell]]
CellGenerator = Iterable[Tuple[Element, Optional[Cell], Optional[CellRange]]]


def union_dicts(dict_1: Dict, dict_2: Dict, with_none_drop: bool = True) -> Dict:
    """
    >>> union_dicts({"a": 1}, {"a": None, "b": 2})
    {'a': 1, 'b': 2}
    """
    if with_none_drop:
        new_dict_2 = {key: value for key, value in dict_2.items() if value is not None}
    else:
        new_dict_2 = dict_2

    return {**dict_1, **new_dict_2}


def try_extract_pixels(pixel_str: Optional[str]) -> Optional[float]:
    """
    >>> try_extract_pixels("100px")
    100.0
    >>> try_extract_pixels("") is None
    True
    """
    if not pixel_str:
        return None

    return float(re.findall(r"(\d+)px", pixel_str)[0])


def parse_cell_value(cell_text: str) -> Any:
    """
    >>> parse_cell_value("") is None
    True
    >>> parse_cell_value("ass")
    'ass'
    >>> parse_cell_value("1")
    1
    >>> parse_cell_value("1.2")
    1.2
    """

    if cell_text == "":
        return None

    if cell_text.isdigit():
        return int(cell_text)

    # float(str) break python for very long non-float string ~_~
    if cell_text.replace(".", "", 1).isdigit():
        return float(cell_text)

    return cell_text


def create_cell_range_str(col_index: int, colspan: int, row_index: int, rowspan: int) -> str:
    start_column = get_column_letter(col_index + 1)
    start_row = row_index + 1
    end_column = get_column_letter(col_index + colspan)
    end_row = row_index + rowspan
    cell_range = f"{start_column}{start_row}:{end_column}{end_row}"
    return cell_range


def width_pixels_to_xlsx_units(pixels: float) -> float:
    return pixels / 7.5


def height_pixels_to_xlsx_units(pixels: float) -> float:
    return pixels * 3 / 4



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jinja2xlsx']

package_data = \
{'': ['*']}

install_requires = \
['cached-property>=1.5,<2.0',
 'jinja2>=3.1,<4.0',
 'openpyxl>=3.0,<4.0',
 'requests-html>=0.10.0,<0.11.0']

extras_require = \
{'pil': ['pillow>=9.4,<10.0']}

setup_kwargs = {
    'name': 'jinja2xlsx',
    'version': '1.1.2',
    'description': 'Create xlsx-tables from html-tables',
    'long_description': '# jinja2xlsx\n\nCreate xlsx-tables from html-tables\n\n## Example\n\nGiven html table str\n\nWhen render html to xlsx\n\nThen result xlsx has table values\n\n```python\nfrom jinja2xlsx import render_xlsx\nfrom openpyxl import Workbook\n\nhtml_str = """<!DOCTYPE html>\n<html lang="en">\n    <head>\n        <meta charset="UTF-8">\n        <title>Simple table</title>\n    </head>\n    <body>\n        <table>\n            <tbody>\n                <tr>\n                    <td>1</td>\n                    <td>2</td>\n                </tr>\n                <tr>\n                    <td>3</td>\n                    <td>4</td>\n                </tr>\n            </tbody>\n        </table>\n    </body>\n</html>"""\n\nworkbook: Workbook = render_xlsx(html_str)\nassert tuple(workbook.active.values) == ((1, 2), (3, 4))\n```\n\n## Installation \n\n```\npip install jinja2xlsx\n```\n\nFor images support:\n\n```\npip install jinja2xlsx[pil]\n```\n\n## Development\n\nInstall dependencies:\n\n```\npoetry install\n```\n\nFor images support:\n\n```\npoetry install -E pil\n```\n\nRun tests and linting:\n\n```\npre-commit run -a\n```\n\nInstall pre-commit hooks:\n\n```\npre-commit install\n```\n\n## Extra\n\n### Publish to PyPI\n\n```shell\npoetry publish --build\n```',
    'author': 'potykion',
    'author_email': 'potykion@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/potykion/jinja2xlsx',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Optional

from openpyxl import Workbook

from jinja2xlsx.config import Config
from jinja2xlsx.parse import Parser
from jinja2xlsx.render import Renderer
from jinja2xlsx.style import Style, Stylist


def render(
    html_str: str,
    default_style: Optional[Style] = None,
    config: Optional[Config] = None,
) -> Workbook:
    renderer = Renderer(
        Parser(html_str),
        Stylist(default_style or Style()),
        config or Config(),
    )
    return renderer()
